package avro

import (
	"bytes"
	"consumerDemo-avro_go/avro/gen"
	"encoding/hex"
	"errors"
	"fmt"
	"github.com/actgardner/gogen-avro/v7/compiler"
	"github.com/actgardner/gogen-avro/v7/vm"
	"github.com/pingcap/parser/charset"
	"strconv"
	"time"
)

/**
包含反序列化avro数据已经解析特定格式的数据的方法
*/

func DeserializeRecord(data []byte, oldData bool) (*gen.Record, error) {
	buf := bytes.NewReader(data)
	record := gen.NewRecord()
	describe, err := compiler.CompileSchemaBytes([]byte(record.Schema()), []byte(record.Schema()))
	if err != nil {
		return nil, err
	}
	if oldData {
		describe, err = compiler.CompileSchemaBytes([]byte(record.SchemaV6()), []byte(record.Schema()))
		if err != nil {
			return nil, err
		}
	}

	err = vm.Eval(buf, describe, record)
	if err != nil {
		return nil, err
	}
	return record, nil
}

func ParseInteger(data *gen.Integer) string {
	return data.Value
}

func ParseCharacter(data *gen.Character) (string, error) {
	// 将数据库中的字符集全部转成utf8给用户
	encoding, _ := charset.Lookup(data.Charset)
	if encoding != nil {
		b, err := encoding.NewDecoder().Bytes(data.Value)
		if err == nil {
			return string(b), nil
		}
	}
	return string(data.Value), errors.New(fmt.Sprintf("unknown charset: %v from data: %v\n", data.Charset, data))
}

func ParseDecimal(data *gen.Decimal) string {
	return data.Value
}

func ParseFloat(data *gen.Float) string {
	return strconv.FormatFloat(data.Value, 'g', -1, 32)
}

func ParseDouble(data *gen.Float) string {
	return strconv.FormatFloat(data.Value, 'g', -1, 64)
}

func ParseTimestamp(data *gen.Timestamp) string {
	t := time.Unix(data.Timestamp, int64(data.Millis*1000000))
	return t.Format("2006-01-02 15:04:05.000")
}

func ParseDateTime(data *gen.DateTime, timezone string, dataTypeNumber int) string {
	tstr := ""
	switch dataTypeNumber {
	case MYSQL_TYPE_DATETIME, MYSQL_TYPE_DATETIME_NEW:
		year := int(data.Year.Int)
		month := int(data.Month.Int)
		day := int(data.Day.Int)
		hour := int(data.Hour.Int)
		minute := int(data.Minute.Int)
		second := int(data.Second.Int)
		milli := 0
		if data.Millis != nil && data.Millis.UnionType == gen.UnionNullIntTypeEnumInt {
			milli = int(data.Millis.Int)
		}
		if len(timezone) > 0 {
			if loc, err := time.LoadLocation(timezone); err == nil {
				datetime := time.Date(year, time.Month(month), day, hour, minute, second, milli, loc)
				tstr = datetime.Format("2006-01-02 15:04:05.000")
			} else {
				fmt.Printf("unknown time zone: %s\n", timezone)
			}
		}
		tstr = fmt.Sprintf("%04d-%02d-%02d %02d:%02d:%02d.%03d", year, month, day, hour, minute, second, milli)
	case MYSQL_TYPE_DATE, MYSQL_TYPE_DATE_NEW:
		year := int(data.Year.Int)
		month := int(data.Month.Int)
		day := int(data.Day.Int)
		tstr = fmt.Sprintf("%04d-%02d-%02d", year, month, day)
	case MYSQL_TYPE_TIME, MYSQL_TYPE_TIME_NEW:
		hour := data.Hour.Int
		minute := data.Minute.Int
		second := data.Second.Int
		milli := 0
		if data.Millis != nil {
			milli = int(data.Millis.Int)
		}
		// 对应mysql time
		tstr = fmt.Sprintf("%02d:%02d:%02d.%03d", hour, minute, second, milli)
	case MYSQL_TYPE_YEAR:
		tstr = strconv.FormatInt(int64(data.Year.Int), 10)
	default:
		fmt.Printf("unsupported time type: %d", dataTypeNumber)
	}
	return tstr
}

func ParseBinaryGeometry(data *gen.BinaryGeometry) string {
	return hex.EncodeToString(data.Value)
}

func ParseTextGeometry(data *gen.TextGeometry) string {
	return data.Value
}

func ParseBinaryObject(data *gen.BinaryObject) string {
	return hex.EncodeToString(data.Value)
}

func ParseTextObject(data *gen.TextObject) string {
	return data.Value
}
